package build

import (
	"git.akyoto.dev/cli/q/src/build/ast"
	"git.akyoto.dev/cli/q/src/build/expression"
	"git.akyoto.dev/cli/q/src/build/keyword"
	"git.akyoto.dev/cli/q/src/build/token"
	"git.akyoto.dev/cli/q/src/errors"
)

// toAST generates an AST from a list of tokens.
func (f *Function) toAST(tokens token.List) (ast.AST, error) {
	tree := make(ast.AST, 0, len(tokens)/64)

	err := EachInstruction(tokens, func(instruction token.List) error {
		node, err := f.toASTNode(instruction)

		if err == nil && node != nil {
			tree = append(tree, node)
		}

		return err
	})

	return tree, err
}

// toASTNode generates an AST node from an instruction.
func (f *Function) toASTNode(tokens token.List) (ast.Node, error) {
	if tokens[0].Kind == token.Keyword {
		switch tokens[0].Text() {
		case keyword.Return:
			value := expression.Parse(tokens[1:])
			return &ast.Return{Value: value}, nil

		case keyword.Loop:
			blockStart := tokens.IndexKind(token.BlockStart) + 1
			blockEnd := tokens.LastIndexKind(token.BlockEnd)

			if blockStart == -1 {
				return nil, errors.New(errors.MissingBlockStart, f.File, tokens[0].End())
			}

			if blockEnd == -1 {
				return nil, errors.New(errors.MissingBlockEnd, f.File, tokens[len(tokens)-1].End())
			}

			tree, err := f.toAST(tokens[blockStart:blockEnd])
			return &ast.Loop{Body: tree}, err

		default:
			return nil, errors.New(&errors.KeywordNotImplemented{Keyword: tokens[0].Text()}, f.File, tokens[0].Position)
		}
	}

	expr := expression.Parse(tokens)

	if expr == nil {
		return nil, nil
	}

	switch {
	case isVariableDefinition(expr):
		if len(expr.Children) < 2 {
			return nil, errors.New(errors.MissingAssignValue, f.File, expr.LastChild().Token.End())
		}

		name := expr.Children[0].Token
		value := expr.Children[1]
		return &ast.Define{Name: name, Value: value}, nil

	case isAssignment(expr):
		if len(expr.Children) < 2 {
			return nil, errors.New(errors.MissingAssignValue, f.File, expr.LastChild().Token.End())
		}

		name := expr.Children[0].Token
		value := expr.Children[1]
		return &ast.Assign{Name: name, Value: value}, nil

	case isFunctionCall(expr):
		return &ast.Call{Expression: expr}, nil

	default:
		return nil, errors.New(&errors.InvalidInstruction{Instruction: expr.Token.Text()}, f.File, expr.Token.Position)
	}
}
